#############################################################################
# 
# This script creates a mosaic from individual raster layers.
# Overlapping extents of adjoining rasters are used to determine
# the gamma correction to apply to the non-reference raster.
#
# To run the script you need to specify the landsat bands for creating
# mosaic images as well as the base directory and folder locations of the individual
# landsat tiles.  When applying the GammaCorrect function across the set
# of rasters, each set of reference and non-reference rasters must be
# specified.
#
# This script was written by Grant Connette [grmcco@gmail.com]
#
#############################################################################

# Function to replace 0 reflectance with NA
NA.fun <- function(x) { x[x==0] <- NA; return(x) }

GammaCorrect <- function(master,tofix,mask){   # master=Reference Raster/tofix=Raster to correct/mask=water,cloud,shadow mask
  new.ext <- intersect(extent(master),extent(tofix))   # Find intersection of the two rasters
  master1 <- crop(master,new.ext)                      # Crop the overlapping area from master raster
  tofix1 <- crop(tofix,new.ext)                        # Crop the overlapping area from raster to correct
  mask1 <- crop(mask,new.ext)
  medlograt <- median(log(master1[which(mask1[]!=1)])/log(tofix1[which(mask1[]!=1)]),na.rm=T)  # Calculate gamma correction
  tofix[] <- tofix[]^medlograt                         # Apply correction to the raster to correct
  return(tofix)
}

# Read in final mask from cloud and shadow masking script - estimating gamma correction factor excludes water pixels
for (TILE in unique(gcr[,2])){
  setwd(file.path(OutFolder, imageDirArray[TILE,1]))
  assign(paste0("mask",".",TILE),raster("FinalMask.tif"))
}  

# Loop over landsat bands
for(j in 2:7){
  cat(paste0("Mosaicking Band ",j),"\n")
  
  # Read in rasters from each tile and assign "TILE.#" naming convention
  for (TILE in 1:nrow(imageDirArray)){
    setwd(file.path(OutFolder, imageDirArray[TILE,1]))
    raster <- paste0("im_B",j,".TIF")
    assign(paste0("TILE",".",TILE),calc(raster(raster),NA.fun))
  }
  
  # Perform gamma corrections specified in "gcr" matrix provided as input  
  for (Correction in 1:nrow(gcr)){
    tofix=get(paste0("TILE",".",gcr[Correction,1]))
    master=get(paste0("TILE",".",gcr[Correction,2]))
    mask=get(paste0("mask",".",gcr[Correction,2]))
    assign(paste0("TILE",".",TILE),GammaCorrect(master=master,tofix=tofix,mask=mask))
  }
  
  # Make a list of all the rasters to be mosaicked
  rlist <- list() # list of rasters
  for (TILE in 1:nrow(imageDirArray)){
    raster <- get(paste0("TILE",".",TILE))
    rlist <- append(rlist, raster) # update raster_list at each iteration
  }
  
  # Specify function to compute cell values in areas where rasters overlap (e.g., mean, min, max, must accept na.rm argument)
  rlist$fun <- mean
  
  # Perform mosaicking on list of rasters and save final raster to output folder with "m_Band#" naming convention
  writeRaster(do.call(mosaic,rlist), filename=paste0(OutFolder,"/m_B",j,".tif"), format="GTiff", overwrite=TRUE)
}

rm(list=ls(pattern="TILE.")); rm(list=ls(pattern="mask.")); gc()  # clear memory
